#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#define BYTE uint8_t
#define WORD uint16_t
#define DWORD uint32_t

/* ------------------------------------------------------------------------- */

#define NUM_SFX_MAX 0x10
#define NUM_CHANNELS    5
#define NUM_FRAMES  0x100
#define MAX_PACKED_LEN  0x100

#define SFXBYTE_NOP 0x7f
#define SFXBYTE_STOP 0x7e

struct sfxdata_raw_s {
    WORD len; /* bytes */
    WORD frames;
    BYTE data[NUM_CHANNELS][NUM_FRAMES];
};

#define SFXDATA_META_CMD_EXTRA_DELAY    0x40
#define SFXDATA_META_CMD_JUMP   0x41
#define SFXDATA_META_CMD_LOOP_COUNTER   0x42
#define SFXDATA_META_CMD_LOOP   0x43
#define SFXDATA_META_CMD_MOD_DELAY  0x44
#define SFXDATA_META_CMD_MOD_ADD    0x45
#define SFXDATA_META_CMD_MOD_EOR    0x46
#define SFXDATA_META_CMD_NUM 7

#define SFX_DEFAULT_VOLUME  8
#define SFX_MASK_CHE  0xf0

struct sfxdata_packed_s {
    WORD len; /* bytes */
    BYTE data[MAX_PACKED_LEN];
};

#define SFXFLAG_NEED_CHA            (1 << 0)
#define SFXFLAG_NEED_CHB            (1 << 1)
#define SFXFLAG_NEED_CHC            (1 << 2)
#define SFXFLAG_NEED_CHD            (1 << 3)
#define SFXFLAG_NEED_CHE            (1 << 4)
#define SFXFLAG_NEED_CH(n)          (1 << ((n) + SFXFLAG_NEED_CHA))
#define SFXFLAG_NEED_EXTRA_DELAY    (1 << 5)
#define SFXFLAG_NEED_JUMP           (1 << 6)
#define SFXFLAG_NEED_LOOP_COUNTER   (1 << 7)
#define SFXFLAG_NEED_LOOP           (1 << 8)
#define SFXFLAG_NEED_MOD_DELAY      (1 << 9)
#define SFXFLAG_NEED_MOD_ADD        (1 << 10)
#define SFXFLAG_NEED_MOD_EOR        (1 << 11)
#define SFXFLAG_NEED_MOD_CHA        (1 << 12)
#define SFXFLAG_NEED_MOD_CHB        (1 << 13)
#define SFXFLAG_NEED_MOD_CHC        (1 << 14)
#define SFXFLAG_NEED_MOD_CHD        (1 << 15)
#define SFXFLAG_NEED_MOD_CHE        (1 << 16)
#define SFXFLAG_NEED_MOD_CH(n)      (1 << ((n) + SFXFLAG_NEED_MOD_CHA))

struct sfxdata_state_s {
    const char *fname[NUM_SFX_MAX];
    struct sfxdata_raw_s raw[NUM_SFX_MAX];
    struct sfxdata_packed_s packed[NUM_SFX_MAX];
    DWORD chflags[NUM_SFX_MAX];
    DWORD gflags;
    BYTE cmdbytes[SFXDATA_META_CMD_NUM];
};

static const char hexchars[16] = "0123456789abcdef";

/* ------------------------------------------------------------------------- */

static int sfxpack_addbyte(const char *fname, struct sfxdata_packed_s *packed, BYTE b)
{
    if (packed->len >= 0x100) {
        fprintf(stderr, "%s: packed data too long!\n", fname);
        return -1;
    }

    packed->data[packed->len++] = b;
    return 0;
}

static int sfxpack_single(const char *fname, struct sfxdata_raw_s *raw, struct sfxdata_packed_s *packed, DWORD *chflags)
{
    DWORD fl = 0;
    BYTE val[NUM_CHANNELS] = { 0, 0, 0, 0, SFX_DEFAULT_VOLUME };
    BYTE b, changed, gotstop = 0;
    int f, fprev = 0, c, delay;
    /* index of previous delay byte (or -1 if none yet) */
    int delay_prev = -1;

    /* TODO detect loops, mod ... */

    for (f = 0; f < raw->frames; ++f) {
        changed = 0;
        for (c = 0; c < NUM_CHANNELS; ++c) {
            b = raw->data[c][f];
            if (b == SFXBYTE_NOP) {
                continue;
            } else if (b == SFXBYTE_STOP) {
                gotstop = 1;
                changed = 0;
                break;
            } else if (b != val[c]) {
                val[c] = b;
                changed |= (1 << (7 - c));
            }
        }
        if (changed || gotstop) {
            /* NEED_CH* flags */
            fl |= changed;

            delay = f - fprev;
            b = (delay > 7) ? 7 : delay;
            if ((b > 0) && (delay_prev >= 0)) {
                packed->data[delay_prev] |= b;
                delay -= b;
            }

            /* TODO extended delay */
            while (delay) {
                b = (delay > 7) ? 7 : delay;
                if (sfxpack_addbyte(fname, packed, b) < 0) {
                    return -1;
                }
                delay -= b;
            }
            delay_prev = packed->len;

            if (gotstop) {
                break;
            }

            if (sfxpack_addbyte(fname, packed, changed) < 0) {
                return -1;
            }

            for (c = 0; c < NUM_CHANNELS; ++c) {
                if (changed & (1 << (7 - c))) {
                    b = val[c];
                    if (c == (NUM_CHANNELS - 1)) {
                        b |= SFX_MASK_CHE;
                    }
                    if (sfxpack_addbyte(fname, packed, b) < 0) {
                        return -1;
                    }
                }
            }
            fprev = f;
        }
    }

    if (sfxpack_addbyte(fname, packed, 0) < 0) {
        return -1;
    }

    *chflags = fl;
    return 0;
}

static void sfxpack_output(int sfxi, struct sfxdata_packed_s *packed)
{
    const BYTE *p = &packed->data[0];
    int left = packed->len;
    int i = 0;
    printf("sfxdata%c:\n", hexchars[sfxi]);
    while (left) {
        if ((i % 16) == 0) {
            if (i > 0) {
                printf("\n");
            }
            printf("!by ");
        } else {
            printf(", ");
        }
        printf("$%02x", *p++);
        ++i, --left;
    }
    printf("\n");
}

static int sfxpack(struct sfxdata_state_s *sfxs, int num_sfx)
{
    int i, c;

    if (num_sfx < NUM_SFX_MAX) {
        ++num_sfx;  /* add "off" sfx */
    }

    for (i = 0; i < num_sfx; ++i) {
        if (sfxpack_single(sfxs->fname[i], &sfxs->raw[i], &sfxs->packed[i], &sfxs->chflags[i]) < 0) {
            return -1;
        }
        sfxs->gflags |= sfxs->chflags[i];
    }

    printf("; ### Generated, do not edit\nSFX_NUM = %i\n", num_sfx);

    for (i = 0; i < num_sfx; ++i) {
        printf("; S%c: %s, $%02x frames, size %i -> %i\n", hexchars[i], sfxs->fname[i] ? sfxs->fname[i] : "(none)", sfxs->raw[i].frames, sfxs->raw[i].len, sfxs->packed[i].len);
    }

    /* output defines */
    printf("; # defines\n");
    for (c = 0; c < NUM_CHANNELS; ++c) {
        printf("SFX_NEED_CH%c = %i\n", c + 'A', SFXFLAG_NEED_CH(c) ? 1 : 0);
    }
    printf("SFX_MASK_DELAY = $07\n");
    printf("SFX_NUM_CHBITS = 5\n");

    /* output (nonempty) sfx */
    printf("; # sfx\n");
    for (i = 0; i < num_sfx; ++i) {
        if (sfxs->packed[i].len > 1) {
            sfxpack_output(i, &sfxs->packed[i]);
        }
    }
    /* empty sfx */
    printf("; # empty sfx (0 byte from first table)\n");
    for (i = 0; i < num_sfx; ++i) {
        if (sfxs->packed[i].len <= 1) {
            printf("sfxdata%c:\n", hexchars[i]);
        }
    }
    /* tables */
    printf("; # tables\n");
    printf("tbl_sfx_offs_chval:\n!by ");
    for (c = 0, i = 0; c < NUM_CHANNELS; ++c) {
        if (c > 0) printf(", ");
        printf("%i", i);
        if (SFXFLAG_NEED_CH(c)) {
            i += 5; /* len(LDA_IMM + STA_ABS) */
        }
    }
    printf("\ntbl_sfx_addr_l:\n!by ");
    for (i = 0; i < num_sfx; ++i) {
        if (i > 0) printf(", ");
        printf("<sfxdata%c", hexchars[i]);
    }
    printf("\ntbl_sfx_addr_h:\n!by ");
    for (i = 0; i < num_sfx; ++i) {
        if (i > 0) printf(", ");
        printf(">sfxdata%c", hexchars[i]);
    }
    printf("\n");
    return 0;
}

/* ------------------------------------------------------------------------- */

static int sfxload(const char *filename, struct sfxdata_raw_s *sr)
{
#define FILELEN (2 + NUM_CHANNELS * NUM_FRAMES)
    FILE *fp = NULL;
    BYTE data[FILELEN];
    long len = 0;
    int c, f;

    sr->len = 0;
    sr->frames = 0;

    if ((fp = fopen(filename, "rb")) == NULL) {
        perror(filename);
        goto fail;
    }

    if (fseek(fp, 0, SEEK_END) != 0) {
        perror(filename);
        goto fail;
    }

    len = ftell(fp);

    if (len > FILELEN) {
        fprintf(stderr, "%s: open, len %li > %i\n", filename, len, FILELEN);
        goto fail;
    }

    rewind(fp);

    if (fread(data, len, 1, fp) < 1) {
        fprintf(stderr, "%s: read error\n", filename);
        goto fail;
    }

    fclose(fp);
    fp = NULL;

    len -= 2;   /* skip load addr */
    sr->len = len;
    sr->frames = len / 5;
    if (len % 5) {
        fprintf(stderr, "%s len mod 5 == %li\n", filename, len % 5);
        ++sr->frames;
    }

    for (c = 0; c < NUM_CHANNELS; ++c) {
        for (f = 0; f < sr->frames; ++f) {
            sr->data[c][f] = data[2 + f * NUM_CHANNELS + c];
        }
    }

    return 0;

fail:
    fclose(fp);
    fp = NULL;
    return -1;
#undef FILELEN
}

/* ------------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    int rc = 0, i, num_files;
    struct sfxdata_state_s sfxs;

    if (argc < 2) {
        fprintf(stderr, "Usage: %s [sfxdataN.prg]+\n", argv[0]);
        return 1;
    }

    memset(&sfxs, 0, sizeof(sfxs));

    --argc;
    ++argv;

    num_files = argc;

    if (num_files > NUM_SFX_MAX) {
        fprintf(stderr, "too many files! max %i\n", NUM_SFX_MAX);
        return 1;
    }

    for (i = 0; i < num_files; ++i) {
        sfxs.fname[i] = argv[i];
        rc = sfxload(sfxs.fname[i], &sfxs.raw[i]);
        if (rc < 0) {
            return -rc;
        }
    }

    rc = sfxpack(&sfxs, num_files);
    return rc;
}
